/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

//* TODO brush feedback

//* TODO for primitives, choose any point or all points

void GroupOp::initialize(void)
{
	catalog<String>("Combine")="";
	catalog<String>("Combine","label")="Input Groups";
	catalog<String>("Combine","page")="Define";
	catalog<bool>("Combine","joined")=true;
	catalog<String>("Combine","hint")=
			"Input group names from which to create a group."
			"  If no names are specified, all elements are considered."
			"  Each space-delimited substring is an individual pattern."
			"  A group qualifies if it matches any of the substring patterns."
			"  Wildcard matching uses globbing unless Regex is turned on.";

	catalog<bool>("Regex")=false;
	catalog<String>("Regex","page")="Define";
	catalog<String>("Regex","hint")=
			"Pass the pattern strings directly as true regular expressions."
			"  If this option is off,"
			" presumed glob patterns are converted to regex."
			"  Every '.' is first escaped as '\\.'."
			"  Then, every '?', indicating 'any character',"
			" is converted to '.'."
			"  Every '*', indicating 'any number of any characters',"
			" is converted to '.*'.";
			"  Every '[!', indicating 'not' inside of square brackets,"
			" is converted to '[^'.";

	catalog<String>("Group")="group";
	catalog<String>("Group","label")="Create Group";
	catalog<String>("Group","page")="Define";
	catalog<String>("Group","hint")=
			"The name of the default group to be created."
			"  If no 'From Attribute' is specified,"
			" this is the only group created.";

	catalog<String>("Attribute")="";
	catalog<String>("Attribute","label")="From Attribute";
	catalog<String>("Attribute","page")="Define";
	catalog<String>("Attribute","hint")=
			"Use the string in the named attribute as the group name"
			" for each element."
			"  If no attribute is specified,"
			" or the attribute does not exist,"
			" or the attribute is empty for an element,"
			" the 'Create Group' name is used.";

	catalog<String>("Element")="Point";
	catalog<String>("Element","page")="Define";
	catalog<String>("Element","choice:0")="Point";
	catalog<String>("Element","choice:1")="Primitive";

	catalog<bool>("EnableBounds")=false;
	catalog<String>("EnableBounds","label")="Enable";
	catalog<String>("EnableBounds","page")="Bounds";
	catalog<bool>("EnableBounds","joined")=true;

	catalog<bool>("Inverse")=false;
	catalog<String>("Inverse","enabler")="EnableBounds";
	catalog<String>("Inverse","page")="Bounds";
	catalog<String>("Inverse","hint")=
		"Include everything that the bounds check would otherwise not include.";

	catalog<String>("Shape")="Box";
	catalog<String>("Shape","enabler")="EnableBounds";
	catalog<String>("Shape","page")="Bounds";
	catalog<String>("Shape","choice:0")="Box";
	catalog<String>("Shape","choice:1")="Sphere";
	catalog<String>("Shape","choice:2")="Surface";

	catalog<SpatialVector>("Center");
	catalog<String>("Center","enabler")="EnableBounds";
	catalog<String>("Center","page")="Bounds";
	catalog<String>("Center","hint")=
			"This is the center of the bounding box or sphere."
			"  For a bounding surface, this will apply a translational"
			" displacement if Displace is turned on.";

	catalog<SpatialVector>("Size")=SpatialVector(1.0,1.0,1.0);
	catalog<String>("Size","enabler")="EnableBounds";
	catalog<String>("Size","page")="Bounds";
	catalog<String>("Size","hint")=
			"This is the dimensions of the bounding box,"
			" the scaling of a unit sphere."
			"  For a bounding surface, this will apply scaling"
			" if Scale is turned on.";

	catalog<bool>("DisplaceSurface")=false;
	catalog<String>("DisplaceSurface","label")="Displace Surface";
	catalog<String>("DisplaceSurface","enabler")="Displace";
	catalog<String>("DisplaceSurface","enabler")="EnableBounds";
	catalog<String>("DisplaceSurface","page")="Bounds";
	catalog<bool>("DisplaceSurface","joined")=true;
	catalog<String>("DisplaceSurface","hint")=
			"This uses Center as a translation when using"
			" an arbitrary bounding surface.";

	catalog<bool>("ScaleSurface")=false;
	catalog<String>("ScaleSurface","label")="Scale Surface";
	catalog<String>("ScaleSurface","enabler")="Displace";
	catalog<String>("ScaleSurface","enabler")="EnableBounds";
	catalog<String>("ScaleSurface","page")="Bounds";
	catalog<String>("ScaleSurface","hint")=
			"This uses Size as scaling when using"
			" an arbitrary bounding surface.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Bounding Surface");
	catalog<bool>("Bounding Surface","optional")=true;
}

void GroupOp::handle(Record& a_rSignal)
{
	const String attribute=catalog<String>("Attribute");
	const BWORD inverse=catalog<bool>("Inverse");
	const SpatialVector center=catalog<SpatialVector>("Center");
	const SpatialVector size=catalog<SpatialVector>("Size");
	const BWORD displaceSurface=catalog<bool>("DisplaceSurface");
	const BWORD scaleSurface=catalog<bool>("ScaleSurface");

	String combine=catalog<String>("Combine");
	if(!catalog<bool>("Regex"))
	{
		combine=combine.convertGlobToRegex();

		feLog("GroupOp::handle converted \"%s\" to \"%s\"\n",
				catalog<String>("Combine").c_str(),combine.c_str());
	}

	m_combined=(!combine.empty());

	String shape=catalog<String>("Shape");

	if(catalog<bool>("EnableBounds"))
	{
		if(shape!="Surface" && (size[0]<=0.0 || size[1]<=0.0 || size[2]<=0.0))
		{
			if(!inverse)
			{
				return;
			}

			//* inverse nothing = everything
			shape="All";
		}
	}
	else
	{
		if(inverse)
		{
			return;
		}

		//* boundless = everything
		shape="All";
	}

	m_doPrimitive=(catalog<String>("Element")=="Primitive");

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(m_combined)
	{
		if(!access(m_spElementPosition,m_spOutputAccessible,
				m_doPrimitive? e_primitiveGroup: e_pointGroup, combine)) return;
	}
	else
	{
		if(!access(m_spElementPosition,m_spOutputAccessible,
				m_doPrimitive? e_primitive: e_point,
				m_doPrimitive? e_vertices: e_position)) return;
	}

	m_spElementAttribute=NULL;
	if(!attribute.empty())
	{
		access(m_spElementAttribute,"Input Surface",
				m_doPrimitive? e_primitive: e_point,attribute,e_warning);
	}

	m_defaultGroup=catalog<String>("Group");
	m_groupMap.clear();

	const U32 elementCount=m_spElementPosition->count();
	if(shape=="Box")
	{
		const SpatialBox box(center-0.5*size,size);
		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			BWORD hit=FALSE;

			const U32 subCount=m_doPrimitive?
					m_spElementPosition->subCount(elementIndex): 1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const SpatialVector point=m_spElementPosition->spatialVector(
						elementIndex,subIndex);

				if((intersecting(box,point)) != inverse)
				{
					hit=TRUE;
					break;
				}
			}
			if(hit)
			{
				groupElement(elementIndex);
			}
		}
	}
	else if(shape=="Sphere")
	{
		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			BWORD hit=FALSE;

			const U32 subCount=m_doPrimitive?
					m_spElementPosition->subCount(elementIndex): 1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const SpatialVector point=m_spElementPosition->spatialVector(
						elementIndex,subIndex);

				const SpatialVector transformed=2.0*SpatialVector(
						(point[0]-center[0])/size[0],
						(point[1]-center[1])/size[1],
						(point[2]-center[2])/size[2]);

				if((magnitudeSquared(transformed)<1.0) != inverse)
				{
					hit=TRUE;
					break;
				}
			}
			if(hit)
			{
				groupElement(elementIndex);
			}
		}
	}
	else if(shape=="Surface")
	{
		sp<SurfaceI> spBounds;
		if(!access(spBounds,"Bounding Surface")) return;

		spBounds->setRefinement(0);

		//* NOTE warn if there are no normals
		sp<SurfaceAccessorI> spBoundNormal;
		access(spBoundNormal,"Bounding Surface",e_point,e_normal,e_warning);

		const SpatialVector reCenter=
				displaceSurface? center: SpatialVector(0.0,0.0,0.0);
		const SpatialVector reSize=
				scaleSurface? size: SpatialVector(1.0,1.0,1.0);

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			BWORD hit=FALSE;

			const U32 subCount=m_doPrimitive?
					m_spElementPosition->subCount(elementIndex): 1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const SpatialVector point=m_spElementPosition->spatialVector(
						elementIndex,subIndex);

				const SpatialVector transformed=SpatialVector(
						(point[0]-reCenter[0])/reSize[0],
						(point[1]-reCenter[1])/reSize[1],
						(point[2]-reCenter[2])/reSize[2]);

				const SurfaceI::Containment containment=
						spBounds->containment(transformed);

				if((containment==SurfaceI::e_inside) != inverse)
				{
					hit=TRUE;
					break;
				}
			}
			if(hit)
			{
				groupElement(elementIndex);
			}
		}
	}
	else
	{
		//* boundless

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			groupElement(elementIndex);
		}
	}
}

void GroupOp::groupElement(I32 a_index)
{
	const I32 reIndex=
			m_combined? m_spElementPosition->integer(a_index): a_index;

	String groupName;
	if(m_spElementAttribute.isValid())
	{
		groupName=m_spElementAttribute->string(reIndex);
		if(groupName.empty())
		{
			groupName=m_defaultGroup;
		}
	}
	else
	{
		groupName=m_defaultGroup;
	}

	if(groupName.empty())
	{
		return;
	}

	sp<SurfaceAccessorI>& rspElementGroup=m_groupMap[groupName];
	if(rspElementGroup.isNull() &&
				!access(rspElementGroup,m_spOutputAccessible,
				m_doPrimitive? e_primitiveGroup: e_pointGroup,groupName))
	{
		return;
	}

	rspElementGroup->append(reIndex);
}

} // namespace
