/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SGO_DEBUG		FALSE

#include "operator/operator.pmh"

using namespace fe;

void HingeOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<Real>("deadRadius")=0.0;
	catalog<Real>("deadRadius","high")=10.0;
	catalog<Real>("deadRadius","max")=1e6;
	catalog<String>("deadRadius","label")="Dead Radius";
	catalog<bool>("deadRadius","joined")=true;

	catalog<Real>("liveRadius")=0.0;
	catalog<Real>("liveRadius","high")=10.0;
	catalog<Real>("liveRadius","max")=1e6;
	catalog<String>("liveRadius","label")="Live Radius";

	catalog<I32>("passes")=1;
	catalog<I32>("passes","min")=1;
	catalog<String>("passes","label")="Passes";

	catalog<I32>("faceA")=0;
	catalog<I32>("faceA","high")=100;
	catalog<I32>("faceA","max")=INT_MAX;
	catalog<String>("faceA","label")="Face A";
	catalog<bool>("faceA","joined")=true;

	catalog<SpatialVector>("baryA")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryA","label")="Barycenter A";

	catalog<SpatialVector>("offsetA");
	catalog<String>("offsetA","label")="Offset A";

	catalog<I32>("faceB")=0;
	catalog<I32>("faceB","high")=100;
	catalog<I32>("faceB","max")=INT_MAX;
	catalog<String>("faceB","label")="Face B";
	catalog<bool>("faceB","joined")=true;

	catalog<SpatialVector>("baryB")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryB","label")="Barycenter B";

	catalog<SpatialVector>("offsetB");
	catalog<String>("offsetB","label")="Offset B";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Collider Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;
}

void HingeOp::handle(Record& a_rSignal)
{
#if FE_SGO_DEBUG
	feLog("HingeOp::handle\n");
#endif

	const Real deadRadius=catalog<Real>("deadRadius");
	const Real liveRadius=catalog<Real>("liveRadius");
	const I32 passes=catalog<I32>("passes");

	sp<DrawI> spDrawGuide;
	accessGuide(spDrawGuide,a_rSignal,e_warning);

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceI> spCollider;
	if(!access(spCollider,"Collider Surface")) return;

	spCollider->setRefinement(0);

	sp<SurfaceI> spDriver;
	access(spDriver,"Driver Surface",e_quiet);

	SpatialVector pivotA=catalog<SpatialVector>("offsetA");
	SpatialVector pivotB=catalog<SpatialVector>("offsetB");
	if(spDriver.isValid())
	{
		const I32 triangleIndexA=catalog<I32>("faceA");
		const SpatialBary baryA=catalog<SpatialVector>("baryA");
		const SpatialTransform xformA=spDriver->sample(triangleIndexA,baryA);
		transformVector(xformA,pivotA,pivotA);

		const I32 triangleIndexB=catalog<I32>("faceB");
		const SpatialBary baryB=catalog<SpatialVector>("baryB");
		const SpatialTransform xformB=spDriver->sample(triangleIndexB,baryB);
		transformVector(xformB,pivotB,pivotB);
	}

	if(spDrawGuide.isValid())
	{
//		sp<SurfaceTriangles> spSurfaceTriangles(spCollider);
//		if(spSurfaceTriangles.isValid())
//		{
//			spSurfaceTriangles->draw(spDrawGuide,NULL);
//		}

		const Color green(0.0,1.0,0.0);
		const Color blue(0.0,0.0,1.0);

		Color colors[2];
		colors[0]=green;
		colors[1]=blue;

		SpatialVector line[2];
		line[0]=pivotA;
		line[1]=pivotB;

		spDrawGuide->drawLines(line,NULL,2,DrawI::e_strip,TRUE,colors);
	}

	const SpatialVector pivot=0.5*(pivotA+pivotB);
	const SpatialVector axis=unitSafe(pivotB-pivotA);

	const U32 pointCount=spOutputPoint->count();

	for(I32 pass=0;pass<passes;pass++)
	{
		Real maxAngle=0.0;

		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=spOutputPoint->spatialVector(pointIndex);

#if FE_SGO_DEBUG
			feLog("\npoint %d/%d %s\n",pointIndex,pointCount,c_print(point));
#endif

			const SpatialVector from=point-pivot;
			const Real radius=magnitude(from);

			if(radius<=0.0 || radius<=deadRadius)
			{
				continue;
			}

			const SpatialVector fromUnit=from*(1.0/radius);
			const SpatialVector fromSide=cross(axis,fromUnit);

			const Real maxDistance= -1.0;
#if FALSE
			sp<SurfaceI::ImpactI> spImpact=
					spCollider->nearestPoint(point,maxDistance);
#else
			const SpatialVector rayDir=unitSafe(fromSide);

#if FE_SGO_DEBUG
			feLog("rayDir %s\n",c_print(rayDir));
#endif

			if(spDrawGuide.isValid())
			{
				const Color black(0.0,0.0,0.0);
				const Color lightred(1.0,0.3,0.3);

				Color colors[2];
				colors[0]=black;
				colors[1]=lightred;

				SpatialVector line[2];
				line[0]=point;
				line[1]=point+rayDir;

				spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);
			}

			sp<SurfaceI::ImpactI> spImpact=
					spCollider->rayImpact(point,rayDir,maxDistance);
#endif
			if(spImpact.isValid())
			{
				const SpatialVector intersection=spImpact->intersection();

#if FE_SGO_DEBUG
				const SpatialVector hitNormal=spImpact->normal();
				feLog("intersect %s normal %s\n",
						c_print(intersection),c_print(hitNormal));
#endif

				const SpatialVector toUnit=unitSafe(intersection-pivot);

				//* TODO not always fully locked to axis
				const SpatialVector fromLocked=unitSafe(cross(fromSide,axis));
				const SpatialVector toSide=cross(axis,toUnit);
				const SpatialVector toLocked=unitSafe(cross(toSide,axis));

#if FE_SGO_DEBUG
				feLog("from %s to %s\n",
						c_print(fromLocked),c_print(toLocked));
#endif

				SpatialQuaternion rotation;
				set(rotation,fromLocked,toLocked);

				Real rotAngle=0.0;
				SpatialVector rotAxis;
				rotation.computeAngleAxis(rotAngle,rotAxis);

				const Real dotAxis=dot(axis,rotAxis);
				if(dotAxis<0.0)
				{
					rotAngle= -rotAngle;
				}

				Real fade=1.0;
				if(radius<liveRadius && liveRadius>deadRadius)
				{
					fade=(radius-deadRadius)/(liveRadius-deadRadius);
				}

#if FE_SGO_DEBUG
				feLog("rotation %s\n",c_print(rotation));
				feLog("angle %.6G axis %s fade %.6G\n",
						rotAngle,c_print(rotAxis),fade);
#endif

				rotAngle*=fade;

				if(maxAngle<rotAngle)
				{
					maxAngle=rotAngle;
				}
			}
		}

		if(maxAngle>0.0)
		{
			const SpatialQuaternion rotation(maxAngle,axis);

			SpatialTransform conversion;
			setIdentity(conversion);
			translate(conversion,pivot);
			conversion*=SpatialTransform(rotation);
			translate(conversion,-pivot);

#if FE_SGO_DEBUG
			feLog("conversion\n%s\n",c_print(conversion));
#endif

			//* rotate all points
			for(U32 otherIndex=0;otherIndex<pointCount;otherIndex++)
			{
				const SpatialVector otherPoint=
						spOutputPoint->spatialVector(otherIndex);

				SpatialVector transformed;
				transformVector(conversion,otherPoint,transformed);

				spOutputPoint->set(otherIndex,transformed);
			}
		}
	}

#if FE_SGO_DEBUG
	feLog("HingeOp::handle done\n");
#endif
}
