/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

#define	FE_NOP_DEBUG	FALSE

namespace fe
{

//* TODO make sure driver has UV data

void SurfaceNormalOp::initialize(void)
{
	catalog<String>("Group");
	catalog<String>("Group","label")="Point Group";
	catalog<String>("Group","suggest")="pointGroups";

	catalog<bool>("PointCustom")=false;
	catalog<String>("PointCustom","label")="Custom Point Attribute";
	catalog<bool>("PointCustom","joined")=true;

	catalog<String>("PointName")="N";
	catalog<String>("PointName","label")="Point Attribute";
	catalog<String>("PointName","enabler")="PointCustom";

	catalog<bool>("PrimitiveCustom")=false;
	catalog<String>("PrimitiveCustom","label")="Custom Primitive Attribute";
	catalog<bool>("PrimitiveCustom","joined")=true;

	catalog<String>("PrimitiveName")="N";
	catalog<String>("PrimitiveName","label")="Primitive Attribute";
	catalog<String>("PrimitiveName","enabler")="PrimitiveCustom";

	catalog<SpatialVector>("DefaultNormal")=SpatialVector(0.0,0.0,1.0);
	catalog<String>("DefaultNormal","label")="Default Normal";
	catalog<String>("DefaultNormal","hint")=
			"On curves without a driver surface or usable bend estimation,"
			"start each curve with this normal.";

	catalog<bool>("PointNormals")=true;
	catalog<String>("PointNormals","label")="Add Point Normals";

	catalog<bool>("FollowBend")=false;
	catalog<String>("FollowBend","label")="Follow Bend";
	catalog<String>("FollowBend","enabler")="PointNormals";
	catalog<String>("FollowBend","hint")=
			"On curves without a driver surface,"
			" try to generate outward point normals"
			" in the same plane as the overall bend of the curve.";

	catalog<bool>("orthogonalize")=true;
	catalog<String>("orthogonalize","label")="Ortho";
	catalog<String>("orthogonalize","enabler")="PointNormals";
	catalog<String>("orthogonalize","hint")=
			"On curves, realign normals to ensure that each point's normal"
			" is perpendicular to either the segment connecting"
			" the adjacent points, or when at the root or tip,"
			" the first or last segment.";

	catalog<bool>("PrimitiveNormals")=false;
	catalog<String>("PrimitiveNormals","label")="Add Primitive Normals";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Input Reference");
	catalog<bool>("Input Reference","optional")=true;

	catalog< sp<Component> >("Driver Reference");
	catalog<bool>("Driver Reference","optional")=true;
}

void SurfaceNormalOp::handle(Record& a_rSignal)
{
	const BWORD pointNormals=catalog<bool>("PointNormals");
	const BWORD primitiveNormals=catalog<bool>("PrimitiveNormals");
	const BWORD followBend=catalog<bool>("FollowBend");
	const BWORD orthogonalize=catalog<bool>("orthogonalize");

	const BWORD pointCustom=catalog<bool>("PointCustom");
	const String pointName=
			pointCustom? catalog<String>("PointName"): "";

	const BWORD primitiveCustom=catalog<bool>("PrimitiveCustom");
	const String primitiveName=
			primitiveCustom? catalog<String>("PrimitiveName"): "";

	const SpatialVector defaultNormal=catalog<SpatialVector>("DefaultNormal");

	if(!pointNormals && !primitiveNormals)
	{
		return;
	}

	const String group=catalog<String>("Group");
	const BWORD grouped=(!group.empty());

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputProperties;
	if(!access(spOutputProperties,spOutputAccessible,
			e_primitive,e_properties)) return;

	sp<SurfaceI> spDriver;
	access(spDriver,"Driver Surface",e_quiet);
	if(spDriver.isValid())
	{
		spDriver->setRefinement(0);
	}

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputGroup;
	if(grouped)
	{
		if(!access(spOutputGroup,spOutputAccessible,
				e_pointGroup,group)) return;
	}
	else
	{
		spOutputGroup=spOutputPoint;
	}

	const U32 primitiveCount=spOutputVertices->count();
	U32 referenceCount=0;

	sp<SurfaceAccessorI> spReferenceVertices;
	sp<SurfaceI> spDriverReference;
	if(access(spReferenceVertices,"Input Reference",
			e_primitive,e_vertices,e_quiet))
	{
		referenceCount=spReferenceVertices->count();
		if(referenceCount!=primitiveCount)
		{
			String message;
			message.sPrintf("Ignoring input reference with different"
					" number of primitives (%d vs %d);",
					referenceCount,primitiveCount);
			catalog<String>("warning")=message;

			spReferenceVertices=NULL;
		}
		else
		{
			access(spDriverReference,"Driver Reference",e_quiet);
			if(spDriverReference.isValid())
			{
				spDriverReference->setRefinement(0);
			}
			else
			{
				spReferenceVertices=NULL;
			}
		}
	}

	const U32 pointCount=spOutputPoint->count();

	const U32 count=spOutputGroup->count();
	if(!count)
	{
		feLog("SurfaceNormalOp::handle point count is zero\n");
		catalog<String>("warning")="no points in input surface";
		return;
	}

#if	FE_NOP_DEBUG
	feLog("SurfaceNormalOp::handle count %d\n",count);
#endif

	sp<SurfaceAccessorI> spPrimitiveNormal;
	if(primitiveNormals)
	{
		if(primitiveCustom)
		{
			if(!access(spPrimitiveNormal,spOutputAccessible,
					e_primitive,primitiveName,e_quiet,e_createMissing)) return;
		}
		else
		{
			if(!access(spPrimitiveNormal,spOutputAccessible,
					e_primitive,e_normal,e_quiet,e_createMissing)) return;
		}
	}

	sp<SurfaceAccessorI> spPointNormal;
	BWORD inGroup[pointCount];
	SpatialVector pointNormal[pointCount];
	if(pointNormals)
	{
		BWORD clearOthers=FALSE;

		if(pointCustom)
		{
			if(!grouped || !access(spPointNormal,spOutputAccessible,
					e_point,pointName,e_quiet))
			{
				clearOthers=grouped;
				if(!access(spPointNormal,spOutputAccessible,e_point,pointName,
						e_quiet,e_createMissing)) return;
			}
		}
		else
		{
			if(!grouped || !access(spPointNormal,spOutputAccessible,
					e_point,e_normal,e_quiet))
			{
				clearOthers=grouped;
				if(!access(spPointNormal,spOutputAccessible,e_point,e_normal,
						e_quiet,e_createMissing)) return;
			}
		}

		const SpatialVector zero(0.0,0.0,0.0);

		if(grouped)
		{
			for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				inGroup[pointIndex]=FALSE;
			}
		}

		//* clear normals
		if(clearOthers)
		{
			for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				set(pointNormal[pointIndex]);
			}
			for(U32 index=0;index<count;index++)
			{
				const I32 pointIndex=spOutputGroup->integer(index);

#if	FE_NOP_DEBUG
				feLog("  %d/%d affects point %d\n",index,count,pointIndex);
#endif

				inGroup[pointIndex]=TRUE;
			}
		}
		else
		{
			for(U32 index=0;index<count;index++)
			{
				const I32 pointIndex=
						grouped? spOutputGroup->integer(index): index;

#if	FE_NOP_DEBUG
				feLog("  %d/%d affects point %d\n",index,count,pointIndex);
#endif

				inGroup[pointIndex]=TRUE;

				set(pointNormal[pointIndex]);
			}
		}
	}

	//* accumulate face normals to points
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spOutputVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		const BWORD openCurve=
				spOutputProperties->integer(primitiveIndex,e_openCurve);
		const U32 countU=
				spOutputProperties->integer(primitiveIndex,e_countU);
		const U32 countV=
				spOutputProperties->integer(primitiveIndex,e_countV);
		const BWORD wrappedU=
				spOutputProperties->integer(primitiveIndex,e_wrappedU);
		const BWORD wrappedV=
				spOutputProperties->integer(primitiveIndex,e_wrappedV);

#if	FE_NOP_DEBUG
		feLog("SurfaceNormalOp::handle primitive %d/%d subCount %d\n",
				primitiveIndex,primitiveCount,subCount);
		feLog("  curve %d mesh %d %d wrapped %d %d\n",
				openCurve,countU,countV,wrappedU,wrappedV);
#endif

		if(openCurve)
		{
			//* open curves

			const SpatialVector rootPoint=
					spOutputVertices->spatialVector(primitiveIndex,0);

			BWORD useSideVector=FALSE;
			SpatialVector sideVector(0.0,1.0,0.0);

			SpatialVector normal=defaultNormal;
			if(spDriver.isValid())
			{
				if(spDriverReference.isValid() && spReferenceVertices.isValid())
				{
					const SpatialVector rootRef=
							spReferenceVertices->spatialVector(
							primitiveIndex,0);

					sp<SurfaceI::ImpactI> spImpactRef=
							spDriverReference->nearestPoint(rootRef);
					if(spImpactRef.isValid())
					{
						const I32 triangleIndex=spImpactRef->triangleIndex();
						const SpatialBary bary=spImpactRef->barycenter();

						sp<SurfaceI::ImpactI> spImpact=
								spDriver->sampleImpact(triangleIndex,bary);
						if(spImpact.isValid())
						{
							normal=spImpact->du();
						}
					}
				}
				else
				{
					sp<SurfaceI::ImpactI> spImpact=
							spDriver->nearestPoint(rootPoint);
					if(spImpact.isValid())
					{
						normal=spImpact->du();
					}
				}
			}


			if(subCount>1)
			{
				const SpatialVector secondPoint=
						spOutputVertices->spatialVector(primitiveIndex,1);
				const SpatialVector firstSegment=secondPoint-rootPoint;

				if(followBend && spDriver.isNull())
				{
					const SpatialVector lastPoint=
							spOutputVertices->spatialVector(
							primitiveIndex,subCount-1);
					const SpatialVector reach=lastPoint-rootPoint;
					sideVector=cross(reach,firstSegment);

					if(magnitudeSquared(sideVector)>1e-6)
					{
						normal=cross(sideVector,firstSegment);
						useSideVector=TRUE;
					}
				}

				if(orthogonalize)
				{
					//* first segment
					const SpatialVector side=cross(firstSegment,normal);
					normal=cross(side,firstSegment);
				}
			}

			if(primitiveNormals)
			{
				spPrimitiveNormal->set(primitiveIndex,unitSafe(normal));
			}
			if(!pointNormals)
			{
				continue;
			}

			const I32 rootIndex=
					spOutputVertices->integer(primitiveIndex,0);
			if(!grouped || inGroup[rootIndex])
			{
//				spPointNormal->set(rootIndex,unitSafe(normal));
				pointNormal[rootIndex]=unitSafe(normal);
			}

			if(subCount<2)
			{
				continue;
			}

			SpatialVector lastPoint=rootPoint;
			SpatialVector lastNormal=normal;
			SpatialVector lastSegment;
			Real lastLength=0.0;

			SpatialVector thisPoint=
					spOutputVertices->spatialVector(primitiveIndex,1);

			for(U32 subIndex=1;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);
				const BWORD writeNormal=(!grouped || inGroup[pointIndex]);

				if(subIndex==subCount-1)
				{
					if(writeNormal)
					{
						normal=lastNormal;
						if(orthogonalize)
						{
							//* last segment
							const SpatialVector side=cross(lastSegment,normal);
							normal=cross(side,lastSegment);
						}
//						spPointNormal->set(pointIndex,unitSafe(normal));
						pointNormal[pointIndex]=unitSafe(normal);
					}
					continue;
				}

				const SpatialVector nextPoint=spOutputVertices->spatialVector(
						primitiveIndex,subIndex+1);

				//* TODO groups

				SpatialVector nextSegment=nextPoint-thisPoint;
				Real nextLength=magnitude(nextSegment);
				nextSegment*=(nextLength>0.0)? 1.0/nextLength: 1.0;

				if(subIndex==1)
				{
					lastSegment=thisPoint-lastPoint;
					lastLength=magnitude(lastSegment);
					lastSegment*=(lastLength>0.0)? 1.0/lastLength: 1.0;
				}

				SpatialTransform rotation;
				set(rotation,lastSegment,nextSegment);

				SpatialVector nextNormal;
				if(useSideVector)
				{
					nextNormal=unitSafe(cross(sideVector,nextSegment));
				}
				else
				{
					rotateVector(rotation,lastNormal,nextNormal);
				}

				normal=lastLength*lastNormal+nextLength*nextNormal;

				if(writeNormal)
				{
					if(orthogonalize)
					{
						//* not first or last segment
						const SpatialVector orthoSegment=nextPoint-lastPoint;
						const SpatialVector side=cross(orthoSegment,normal);
						normal=cross(side,orthoSegment);
					}

//					spPointNormal->set(pointIndex,unitSafe(normal));
					pointNormal[pointIndex]=unitSafe(normal);
				}

				lastPoint=thisPoint;
				lastNormal=nextNormal;
				lastSegment=nextSegment;
				lastLength=nextLength;

				thisPoint=nextPoint;
			}

			continue;
		}

		//* mesh
		if(countV>1)
		{
			const U32 maxU=wrappedU? countU: countU-1;
			const U32 maxV=wrappedV? countV: countV-1;

			for(U32 v=0;v<maxV;v++)
			{
				const U32 vv=v*countU;
				for(U32 u=0;u<maxU;u++)
				{
					U32 p[4];
					p[0]=vv+u;
					p[1]=(u==countU-1)? vv: p[0]+1;
					p[2]=(v==countV-1)? u: p[0]+countU;
					p[3]=(u==countU-1)? p[2]-(countU-1): p[2]+1;

					SpatialVector sv[4];
					sv[0]=spOutputVertices->spatialVector(primitiveIndex,p[0]);
					sv[1]=spOutputVertices->spatialVector(primitiveIndex,p[1]);
					sv[2]=spOutputVertices->spatialVector(primitiveIndex,p[2]);
					sv[3]=spOutputVertices->spatialVector(primitiveIndex,p[3]);

					SpatialVector faceNormal[2]=
					faceNormal[0]= -unitSafe(cross(sv[1]-sv[0],sv[2]-sv[0]));
					faceNormal[1]= -unitSafe(cross(sv[1]-sv[2],sv[3]-sv[2]));

					if(pointNormals)
					{
						for(U32 half=0;half<2;half++)
						{
							for(U32 m=0;m<3;m++)
							{
								const U32 mm=half+m;
								const I32 pointIndex=spOutputVertices->integer(
										primitiveIndex,p[mm]);
								if(!grouped || inGroup[pointIndex])
								{
//									const SpatialVector normal=
//											spPointNormal->spatialVector(
//											pointIndex);
//									spPointNormal->set(pointIndex,
//											normal+faceNormal[half]);
									pointNormal[pointIndex]+=faceNormal[half];
#if	FE_NOP_DEBUG
									feLog("    uv %d/%d %d/%d"
											" half %d n %s\n",
											u,countU,v,countV,half,
											c_print(pointNormal[pointIndex]));
#endif
								}
							}
						}
					}

					if(primitiveNormals)
					{
						const SpatialVector normal=
								spPrimitiveNormal->spatialVector(
								primitiveIndex);
						spPrimitiveNormal->set(primitiveIndex,
								normal+faceNormal[0]+faceNormal[1]);
					}
				}
			}
			if(primitiveNormals)
			{
				const SpatialVector normal=
						spPrimitiveNormal->spatialVector(primitiveIndex);
				spPrimitiveNormal->set(primitiveIndex,unitSafe(normal));
			}
			continue;
		}

		if(subCount<3)
		{
			continue;
		}

		//* otherwise, presume triangles

		const SpatialVector p0=
				spOutputVertices->spatialVector(primitiveIndex,0);
		SpatialVector p1;
		SpatialVector p2;

		if(subCount==3)
		{
			p1=spOutputVertices->spatialVector(primitiveIndex,1);
			p2=spOutputVertices->spatialVector(primitiveIndex,2);
		}
		else
		{
			//* NOTE presuming quads
			const I32 pointIndex0=
					spOutputVertices->integer(primitiveIndex,0);
			I32 pointIndex1=spOutputVertices->integer(primitiveIndex,1);
			if(pointIndex1==pointIndex0)
			{
				p1=spOutputVertices->spatialVector(primitiveIndex,2);
				p2=spOutputVertices->spatialVector(primitiveIndex,3);
			}
			else
			{
				I32 pointIndex2=spOutputVertices->integer(primitiveIndex,2);
				if(pointIndex2==pointIndex0 || pointIndex2==pointIndex1)
				{
					p1=spOutputVertices->spatialVector(primitiveIndex,1);
					p2=spOutputVertices->spatialVector(primitiveIndex,3);
				}
				else
				{
					p1=spOutputVertices->spatialVector(primitiveIndex,1);
					p2=spOutputVertices->spatialVector(primitiveIndex,2);
				}
			}
		}


		const SpatialVector faceNormal= -unitSafe(cross(p1-p0,p2-p0));

#if	FE_NOP_DEBUG
		const I32 v0=spOutputVertices->integer(primitiveIndex,0);
		const I32 v1=spOutputVertices->integer(primitiveIndex,1);
		const I32 v2=spOutputVertices->integer(primitiveIndex,2);
		feLog("  prim %d/%d v0 %d v1 %d v2 %d grouped %d %d %d\n",
					primitiveIndex,primitiveCount,v0,v1,v2,
					inGroup[v0],inGroup[v1],inGroup[v2]);
		feLog("  p0 %s p1 %s p2 %s n %s\n",
					c_print(p0),c_print(p1),c_print(p2),c_print(faceNormal));
#endif

		if(pointNormals)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spOutputVertices->integer(primitiveIndex,subIndex);

				if(!grouped || inGroup[pointIndex])
				{
//					const SpatialVector normal=
//							spPointNormal->spatialVector(pointIndex);
//					spPointNormal->set(pointIndex,normal+faceNormal);
					pointNormal[pointIndex]+=faceNormal;
#if	FE_NOP_DEBUG
					feLog("    vert %d/%d pt %d n %s\n",
							subIndex,subCount,pointIndex,
							c_print(pointNormal[pointIndex]));
#endif
				}
			}
		}

		if(primitiveNormals)
		{
			spPrimitiveNormal->set(primitiveIndex,faceNormal);
		}
	}

	//* normalize accumulated normals
	if(pointNormals)
	{
		for(U32 index=0;index<count;index++)
		{
			const I32 pointIndex=grouped? spOutputGroup->integer(index): index;

//			const SpatialVector normal=
//					spPointNormal->spatialVector(pointIndex);
//			spPointNormal->set(pointIndex,unitSafe(normal));
			spPointNormal->set(pointIndex,unitSafe(pointNormal[pointIndex]));
		}
	}
}

} // namespace
