/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

// CORAL

#include "platform.pmh"

#define FESTRING_SUBST_DEBUG	FESTRING_DEBUG

namespace fe
{

void String::Rep::newBuffer(U32 size)
{
	//* for U8 buffer, allocate double the chars for largest possible multibyte

	if(gs_pAllocateFunction)
	{
		m_pBuffer=(FESTRING_U8 *)allocate(2*size*sizeof(FESTRING_U8));
	}
	else
	{
		FEASSERT(0);
		m_pBuffer=(FESTRING_U8 *)malloc(2*size*sizeof(FESTRING_U8));
	}
}

void String::Rep::deleteBuffer(void)
{
	deallocateFunction *pDeallocateFunction=
			gs_pAllocateFunction? gs_pDeallocateFunction: free;

	FEASSERT(pDeallocateFunction);
	if (m_pBuffer)
	{
		pDeallocateFunction(m_pBuffer);
		m_pBuffer = NULL;
	}
}

String::Rep *String::newRep(void)
{
	Rep *pRep=NULL;

	if(gs_pAllocateFunction)
	{
		pRep=(Rep *)allocate(sizeof(Rep));
	}
	else
	{
		FEASSERT(0);
		pRep=(Rep *)malloc(sizeof(Rep));
	}

	pRep->m_references=1;

	return pRep;
}

void String::deleteRep(void)
{
	deallocateFunction *pDeallocateFunction=
			gs_pAllocateFunction? gs_pDeallocateFunction: free;

	m_pRep->deleteBuffer();

	FEASSERT(pDeallocateFunction);
	pDeallocateFunction(m_pRep);

	m_pRep=NULL;
}

String String::pathname(void) const
{
	const FESTRING_U8* str1=rawU8();
	const U32 len1=length();
	if(len1==0)
	{
		return *this;
	}
	const FESTRING_U8* str2=(const FESTRING_U8*)FEMEMRCHR(str1,'/',len1);
	if(!str2 || str2==str1)
	{
		return String();
	}
	FESTRING_U8* buffer=new FESTRING_U8[len1+1];
	memcpy(buffer,str1,len1);
	buffer[str2-str1]=0;
	String result(buffer);
	delete[] buffer;
	return result;
}

String String::basename(void) const
{
	const FESTRING_U8* str1=rawU8();
	const U32 len1=length();
	if(len1==0)
	{
		return *this;
	}
	const FESTRING_U8* str2=(const FESTRING_U8*)FEMEMRCHR(str1,'/',len1);
	if(!str2)
	{
		return *this;
	}
	if(++str2==str1+len1)
	{
		return String();
	}
	return String(str2);
}

String String::prechop(const String& prefix) const
{
	const FESTRING_U8* str1=rawU8();
	const FESTRING_U8* str2=prefix.rawU8();

	const U32 len1=length();
	const U32 len2=prefix.length();

	if(len1<len2 || FESTRNCMP_MB(str1,str2,len2))
	{
		return *this;
	}
	FESTRING_U8* buffer=new FESTRING_U8[len1-len2+1];
	memcpy(buffer,str1+len2,len1-len2);
	buffer[len1-len2]=0;
	String result(buffer);
	delete[] buffer;
	return result;
}

String String::chop(const String& suffix) const
{
	const FESTRING_U8* str1=rawU8();
	const FESTRING_U8* str2=suffix.rawU8();

	const U32 len1=length();
	const U32 len2=suffix.length();

	if(len1<len2 || FESTRNCMP_MB(str1+len1-len2,str2,len2))
	{
		return *this;
	}
	FESTRING_U8* buffer=new FESTRING_U8[len1-len2+1];
	memcpy(buffer,str1,len1-len2);
	buffer[len1-len2]=0;
	String result(buffer);
	delete[] buffer;
	return result;
}

String String::substitute(String find,String replace) const
{
	String result;

//	printf("String::substitute \"%s\" \"%s\" \"%s\"\n",
//			c_str(),find.c_str(),replace.c_str());

	const char* search=find.c_str();
	I32 size=strlen(search);
	const char* replacement=replace.c_str();
	const char* start=c_str();
	char* buffer=new char[strlen(start)];
	while(start)
	{
		const char* end=strstr(start,search);
		if(end)
		{
			strncpy(buffer,start,end-start);
			buffer[end-start]=0;
			result.cat(buffer);
			result.cat(replacement);

//			printf("String::substitute \"%s\" -> \"%s\"\n",
//					start,result.c_str());
			start=end+size;
		}
		else
		{
			result.cat(start);
			start=NULL;
		}
	}
	delete[] buffer;

//	printf("String::substitute \"%s\"\n",result.c_str());

	return result;
}

String String::substituteEnv(void) const
{
	String result;
#if FESTRING_SUBST_DEBUG
	feLog("String::substituteEnv \"%s\"\n",c_str());
#endif

	const I32 maxlength=1024;
	char substring[maxlength];
	const char* string=c_str();
	const char* last=string+strlen(string)-1;
	const char* c;
	while( (c=strchr(string,'$')) )
	{
		I32 length2=c-string;
		if(length2>maxlength)
		{
			length2=maxlength;
		}
		if(length2)
		{
			strncpy(substring,string,length2);
			substring[length2]=0;
#if FESTRING_SUBST_DEBUG
			feLog("String::substituteEnv %s\n",substring);
#endif
			result+=substring;
		}
		string=c;
		if(c+1 < last)
		{
			BWORD bracket=(*(c+1) == '{');
			const char *c2;
			if(bracket)
			{
				c2=strchr(string,'}');
			}
			else
			{
				c2=strchr(string,' ');
				const char* c3=strchr(string,'\t');
				if(c3 && (!c2 || c3<c2))
				{
					c2=c3;
				}
				c3=strchr(string,'\n');
				if(c3 && (!c2 || c3<c2))
				{
					c2=c3;
				}
				if(!c2)
				{
					c2=string+strlen(string);
				}
			}
			if(c2)
			{
				length2=c2-c-1-bracket;
				if(length2<maxlength)
				{
					strncpy(substring,c+1+bracket,length2);
					substring[length2]=0;
					const char *value=getenv(substring);
#if FESTRING_SUBST_DEBUG
					feLog("String::substituteEnv variable %s=%s bracket=%d\n",
							substring,value,bracket);
#endif
					result+=value;
					string=c+length2+(bracket? 3: 1);
				}
			}
		}
		if(string==c)
		{
			result+="$";
			string++;
		}
	}
#if FESTRING_SUBST_DEBUG
	feLog("String::substituteEnv %s\n",string);
#endif
	result+=string;

	return result;
}

BWORD String::dotMatch(const String& operand) const
{
//	printf("dotMatch() \"%s\" \"%s\"\n",rawU8(),operand.rawU8());

	const char* ostr1=c_str();
	const char* ostr2=operand.c_str();

	const I32 olen1=length();
	const I32 olen2=operand.length();

	const char* end1=ostr1+olen1;
	const char* end2=ostr2+olen2;

#if FALSE
	U32 len1=olen1;
	U32 len2=olen2;
	while(str2[0]=='*' && str2[1]=='.')
	{
		const FESTRING_U8* str=(const FESTRING_U8*)memchr(str1,'.',len1);
		if(!str)
		{
			return false;
		}
		str++;
		str2+=2;
		len1-=str-str1;
		len2-=2;
		str1=str;
		FEASSERT(str1<rawU8()+olen1);
		FEASSERT(str2<operand.rawU8()+olen2);

		printf("  rematch \"%s\" \"%s\"\n",str1,str2);
	}

	if(FESTRLEN(str1)>=len2 && (str1[len2]==0 || str1[len2]=='.') &&
			!FESTRNCMP_MB(str1,str2,len2))
	{
		printf("  matched\n");
		return true;
	}

	printf("  not matched\n");
	return false;
#else
	const char* str1=ostr1;
	const char* str2=ostr2;
	I32 len1=olen1;
	I32 len2=olen2;
	while(len1 && len2)
	{
		const char* token1=strchr((const char*)str1,'.');
		const char* token2=strchr((const char*)str2,'.');

		I32 sublen1=token1? token1-str1: len1;
		I32 sublen2=token2? token2-str2: len2;

//		printf("  sub %s:%d %s:%d\n",str1,sublen1,str2,sublen2);

		if(!(sublen2==1 && *str2=='*') &&
				(sublen1!=sublen2 || strncmp(str1,str2,sublen1)))
		{
//			printf("  not matched\n");
			return false;
		}

		len1-=sublen1+(token1!=0);
		len2-=sublen2+(token2!=0);

		str1=end1-len1;
		str2=end2-len2;

		FEASSERT(str1<=end1);
		FEASSERT(str2<=end2);
	}

//	printf("  matched\n");
	return !len2;
#endif
}

String &String::sPrintf(const char* fmt, ...)
{
	int size=0;
	while(size>=0)
	{
		va_list ap;
		va_start(ap, fmt);

		vsPrintf(fmt,ap,size);

		va_end(ap);
	}

	return *this;
}

// Brute force.  Rewrite if necessary for speed or space.
String &String::cat(const char* fmt, ...)
{
	String tmp;

	int size=0;
	while(size>=0)
	{
		va_list ap;
		va_start(ap, fmt);

		tmp.vsPrintf(fmt,ap,size);

		va_end(ap);
	}

	String previous(*this);
	sPrintf("%s%s",previous.c_str(),tmp.c_str());

	return *this;
}

// TODO << operator 
String &String::cat(std::list<String> &strList, String sep)
{
	for(std::list<String>::iterator it = strList.begin();
		it != strList.end(); it++)
	{
		if(it != strList.begin())
		{
			cat("%s", sep.c_str());
		}
		cat("%s", it->c_str());
	}
	return *this;
}

String String::maybeQuote(String quote) const
{
	if(strcspn(c_str()," \t\n") == strlen(c_str()))
	{
		return *this;
	}
	return quote + *this + quote;
}

String String::parse(String quote, String delimiters)
{
	const char* start=c_str()+strspn(c_str(),delimiters.c_str());
	const char* limit=c_str()+strlen(c_str());
	U32 len;
	U32 quotes=0;
	if(strchr(quote.c_str(),start[0]))
	{
		start++;
		len=strcspn(start,quote.c_str());
		quotes=(start+len<limit);
	}
	else
	{
		len=strcspn(start,delimiters.c_str());
	}

	char* copy=new char[strlen(start)+1];
	strcpy(copy,start);
	*this=copy+len+quotes;

//	feLog("\nstart \"%s\"\n",copy);
	copy[len]=0;
//	feLog("token \"%s\"\n",copy);
//	feLog("new   \"%s\"\n",copy+len+quotes);
//	feLog("len=%d\n",len);

	String token=copy;
	delete[] copy;

	return token;
}

String String::join(const String &a_sep, std::list<String> &a_list)
{
	String rv;
	rv.cat(a_list, a_sep);
	return rv;
}

String String::join(const String &a_sep, const String &a_1, const String &a_2)
{
	std::list<String> list;
	list.push_back(a_1);
	list.push_back(a_2);
	String rv;
	rv.cat(list, a_sep);
	return rv;
}

String String::join(const String &a_sep, const String &a_1, const String &a_2,
		const String &a_3)
{
	std::list<String> list;
	list.push_back(a_1);
	list.push_back(a_2);
	list.push_back(a_3);
	String rv;
	rv.cat(list, a_sep);
	return rv;
}

// Brute force.  Rewrite if necessary for speed.
String& String::vsPrintf(const char* fmt, va_list ap,int& size)
{
	if(size<0)
	{
		feX("String::vsPrintf size<0 indicates unheeded discontinuation");
	}
	if(!size)
	{
		// start at 80 bytes, seems reasonable (1 tty line)
		size=80;
	}
	FESTRING_U8 *tmpbuf = new FESTRING_U8[size];
	FEASSERT(NULL != tmpbuf);

	int n=FEVSNPRINTF((char *)tmpbuf, size, fmt, ap);

	if(n> -1 && n<size)
	{
		size= -1;
		operator=(tmpbuf);
		delete [] tmpbuf;
		return *this;
	}
	if(n> -1)
	{
		size=n+1;
	}
	else
	{
		size*=2;
	}

	delete [] tmpbuf;
	operator=("");
	return *this;
}

#if FALSE
String &String::vsPrintf(const char* fmt, va_list ap)
{
	feLog("String::vsPrintf \"%s\"\n",fmt);

	// start at 80 bytes, seems reasonable (1 tty line)
	int sz = 80;
	FESTRING_U8 *tmpbuf = new FESTRING_U8[sz];
	FEASSERT(NULL != tmpbuf);
	while(true)
	{
		int n=FEVSNPRINTF((char *)tmpbuf, sz, fmt, ap);

		if(n> -1 && n<sz)
			break;
		if(n> -1)
			sz=n+1;
		else
			sz*=2;

		delete [] tmpbuf;
		tmpbuf = new FESTRING_U8[sz];
		FEASSERT(NULL != tmpbuf);

		// NOTE calling vsnprintf again without va_end,va_start is undefined
		feX("String::vsPrintf warning, expanding buffer\n");
	}
	operator=(tmpbuf);
	delete [] tmpbuf;

	return *this;
}
#endif

void String::output(std::ostream &ostrm) const
{
	U32 len = FESTRLEN(c_str());
	U32 nlen = htonl(len);
	ostrm.write((char *)&nlen, 4);
	ostrm.write(c_str(), len);
}

void String::input(std::istream &istrm)
{
	U32 len;
	istrm.read((char *)&len, 4);
	len = ntohl(len);

	// TODO remove this extra memory buffer.  input directly into raw buffer
	char *buffer = new char[len + 1];
	istrm.read(buffer, len);
	buffer[len] = 0;
	operator=(buffer);
	delete[] buffer;
}

//* TODO optimize array usage
void String::forceCase(BWORD upper)
{
	if(m_pRep)
	{
		String temp=*this;
		*this=temp.c_str();	// replicate
	}

#ifdef FESTR_MB2UPPER
	if(upper)
		FESTR_MB2UPPER(c_str());
	else
		FESTR_MB2LOWER(c_str());
#else
	//* manual 8-bit version
	U32 m,len=FESTRLEN(c_str());
	for(m=0;m<len;m++)
//		if(m_pRep->m_pBuffer[m] < 0x7f)
			buffer()[m]=upper? toupper(c_str()[m]): tolower(c_str()[m]);
#endif
}

String errorString(int error)
{
	String message;
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	// from MSDN example
	LPVOID lpMsgBuf;
	if (!FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			error,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL ))
		return "no error string";

	message = (char *)lpMsgBuf;

	LocalFree( lpMsgBuf );
#else
	message = strerror(error);
#endif
	return message;
}

int String::integer(void) const
{
	return atoi(c_str());
}

Real String::real(void) const
{
	return (Real)atof(c_str());
}

bool String::match(const String a_pattern) const
{
	return Regex(a_pattern.c_str()).match(c_str());
}

bool String::search(const String a_pattern) const
{
	return Regex(a_pattern.c_str()).search(c_str());
}

String String::replace(const String a_pattern,const String a_replacement) const
{
	return Regex(a_pattern.c_str()).replace(c_str(),a_replacement.c_str());
}

String String::convertGlobToRegex(void) const
{
	String result=*this;

	result=result.substitute(".","\\.");
	result=result.substitute("?",".");
	result=result.substitute("*",".*");
	result=result.substitute("[!",".[^");

	return result;
}

} /* namespace */
